# -*- coding: utf-8 -*-
#
# tabu documentation build configuration file
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# Note: make sure the package is installed when building docs.

# -- Project information -----------------------------------------------------

from tabu import package_info
project = 'D-Wave Tabu'
copyright = '2018, D-Wave Systems Inc.'
author = package_info.__author__
version = package_info.__version__
release = package_info.__version__


# -- General configuration ------------------------------------------------
# import sphinx
# if sphinx.__version__  # can check here

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'recommonmark',
    'sphinx.ext.ifconfig',
]

autosummary_generate = True

# templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
#source_parsers = {'.md': 'recommonmark.parser.CommonMarkParser'}

# The master toctree document.
master_doc = 'index'


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None


add_module_names = False
# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'sdk_index.rst']

linkcheck_retries = 2
linkcheck_anchors = False
linkcheck_ignore = ['https://cloud.dwavesys.com/leap',  # redirects, many checks
                    ]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

modindex_common_prefix = ['dwave-tabu.']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
import sphinx_rtd_theme
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']

def setup(app):
   app.add_css_file('cookie_notice.css')
   app.add_js_file('cookie_notice.js')
   app.add_config_value('target', 'repo', 'env')

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'qbsolv': ('https://docs.ocean.dwavesys.com/projects/qbsolv/en/stable/', None),
    'oceandocs': ('https://docs.ocean.dwavesys.com/en/stable/', None),
    'sysdocs_gettingstarted': ('https://docs.dwavesys.com/docs/stable/', None),
}



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the TabuSampler python interface."""

import unittest

import dimod
import numpy as np

import tabu
from tabu.utils import tictoc


@dimod.testing.load_sampler_bqm_tests(tabu.TabuSampler)
class TestTabuSampler(unittest.TestCase):

    def test_instantiation(self):
        sampler = tabu.TabuSampler()
        dimod.testing.assert_sampler_api(sampler)

    def test_sample_basic(self):
        sampler = tabu.TabuSampler()

        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': +1, 'ac': +1})

        resp = sampler.sample(bqm)

        dimod.testing.assert_response_energies(resp, bqm)

    def test_sample_num_reads(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': +1, 'ac': +1})

        resp = tabu.TabuSampler().sample(bqm, num_reads=57)
        dimod.testing.assert_response_energies(resp, bqm)
        self.assertEqual(sum(resp.record.num_occurrences), 57)

    def test_disconnected_problem(self):
        h = {}
        J = {
            # K_3
            (0, 1): -1,
            (1, 2): -1,
            (0, 2): -1,

            # disonnected K_3
            (3, 4): -1,
            (4, 5): -1,
            (3, 5): -1,
        }

        bqm = dimod.BinaryQuadraticModel.from_ising(h, J)
        resp = tabu.TabuSampler().sample(bqm)
        dimod.testing.assert_response_energies(resp, bqm)

    def test_empty(self):
        resp = tabu.TabuSampler().sample(dimod.BinaryQuadraticModel.empty(dimod.SPIN))
        dimod.testing.assert_response_energies(resp, dimod.BinaryQuadraticModel.empty(dimod.SPIN))

        resp = tabu.TabuSampler().sample(dimod.BinaryQuadraticModel.empty(dimod.BINARY))
        dimod.testing.assert_response_energies(resp, dimod.BinaryQuadraticModel.empty(dimod.BINARY))

        resp = tabu.TabuSampler().sample_qubo({})
        dimod.testing.assert_response_energies(resp, dimod.BinaryQuadraticModel.empty(dimod.BINARY))

        resp = tabu.TabuSampler().sample_ising({}, {})
        dimod.testing.assert_response_energies(resp, dimod.BinaryQuadraticModel.empty(dimod.SPIN))

    def test_single_variable_problem(self):
        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0.0, dimod.SPIN)
        resp = tabu.TabuSampler().sample(bqm)
        dimod.testing.assert_response_energies(resp, bqm)
        self.assertEqual(resp.first.energy, -1)

    def test_linear_problem(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({v: -1 for v in range(100)}, {})
        resp = tabu.TabuSampler().sample(bqm)
        dimod.testing.assert_response_energies(resp, bqm)

    def test_initial_states_smoketest(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})
        resp = tabu.TabuSampler().sample(bqm, initial_states=tabu.TabuSampler().sample(bqm))
        dimod.testing.assert_response_energies(resp, bqm)

    def test_initial_states(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})
        init = dimod.SampleSet.from_samples({'a': 0, 'b': 0, 'c': 0}, vartype=dimod.BINARY, energy=0)

        resp = tabu.TabuSampler().sample(bqm, initial_states=init)
        dimod.testing.assert_response_energies(resp, bqm)

    def test_initial_states_generator(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})
        init = dimod.SampleSet.from_samples_bqm([{'a': 1, 'b': 1, 'c': 1},
                                                 {'a': -1, 'b': -1, 'c': -1}], bqm)

        # 2 fixed initial state, 8 random
        resp = tabu.TabuSampler().sample(bqm, initial_states=init, num_reads=10)
        self.assertEqual(len(resp), 10)

        # 2 fixed initial states, 8 random, explicit
        resp = tabu.TabuSampler().sample(bqm, initial_states=init, initial_states_generator='random', num_reads=10)
        self.assertEqual(len(resp), 10)

        # all random
        resp = tabu.TabuSampler().sample(bqm, initial_states_generator='random', num_reads=10)
        self.assertEqual(len(resp), 10)

        # all random
        resp = tabu.TabuSampler().sample(bqm, num_reads=10)
        self.assertEqual(len(resp), 10)


        # initial_states truncated to num_reads?
        resp = tabu.TabuSampler().sample(bqm, initial_states=init, initial_states_generator='none', num_reads=1)
        self.assertEqual(len(resp), 1)

        resp = tabu.TabuSampler().sample(bqm, initial_states=init, initial_states_generator='tile', num_reads=1)
        self.assertEqual(len(resp), 1)

        resp = tabu.TabuSampler().sample(bqm, initial_states=init, initial_states_generator='random', num_reads=1)
        self.assertEqual(len(resp), 1)


        # 2 fixed initial states, repeated 5 times
        resp = tabu.TabuSampler().sample(bqm, initial_states=init, initial_states_generator='tile', num_reads=10)
        self.assertEqual(len(resp), 10)

        # can't tile empty states
        with self.assertRaises(ValueError):
            resp = tabu.TabuSampler().sample(bqm, initial_states_generator='tile', num_reads=10)

        # not enough initial states
        with self.assertRaises(ValueError):
            resp = tabu.TabuSampler().sample(bqm, initial_states_generator='none', num_reads=3)

        # initial_states incompatible with the bqm
        init = dimod.SampleSet.from_samples({'a': 1, 'b': 1}, vartype='SPIN', energy=0)
        with self.assertRaises(ValueError):
            resp = tabu.TabuSampler().sample(bqm, initial_states=init)

    def test_input_validation(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})
        empty = dimod.BinaryQuadraticModel.empty(dimod.SPIN)

        # empty bqm
        self.assertEqual(len(tabu.TabuSampler().sample(empty)), 0)

        # invalid tenure type
        with self.assertRaises(TypeError):
            tabu.TabuSampler().sample(bqm, tenure=2.0)

        # invalid tenure range
        with self.assertRaises(ValueError):
            tabu.TabuSampler().sample(bqm, tenure=100)

        # invalid num_reads type
        with self.assertRaises(TypeError):
            tabu.TabuSampler().sample(bqm, num_reads=10.0)

        # invalid num_reads range
        with self.assertRaises(ValueError):
            tabu.TabuSampler().sample(bqm, num_reads=0)

        # invalid initial_states type
        with self.assertRaises(ValueError):
            tabu.TabuSampler().sample(bqm, initial_states=[])

        with self.assertRaises(ValueError):
            tabu.TabuSampler().sample(bqm, initial_states_generator='non-existing')

        # invalid initial_states length
        with self.assertRaises(ValueError):
            tabu.TabuSampler().sample(bqm, initial_states=[1, 1])

    def test_soft_num_reads(self):
        """Number of reads adapts to initial_states size, if provided."""

        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})
        init = dimod.SampleSet.from_samples_bqm([{'a': 1, 'b': 1, 'c': 1},
                                                 {'a': -1, 'b': -1, 'c': -1}], bqm)

        # default num_reads == 1
        self.assertEqual(len(tabu.TabuSampler().sample(bqm)), 1)

        # with initial_states, num_reads == len(initial_states)
        self.assertEqual(len(tabu.TabuSampler().sample(bqm, initial_states=init)), 2)

        # if explicitly given, with initial_states, they are expanded
        self.assertEqual(len(tabu.TabuSampler().sample(bqm, initial_states=init, num_reads=3)), 3)

        # if explicitly given, without initial_states, they are generated
        self.assertEqual(len(tabu.TabuSampler().sample(bqm, num_reads=4)), 4)

    def test_seed(self):
        sampler = tabu.TabuSampler()

        # use a bqm with a huge plateau (extreme ground state degeneracy)
        # to give the tabu sampler many chances of flipping (exercising rng)
        bqm = dimod.generators.random.randint(10, dimod.SPIN, low=1, high=1)
        tenure = 5

        all_samples = []
        for seed in (1, 25, 2352):
            response0 = sampler.sample(bqm, num_reads=1, tenure=tenure, num_restarts=1, timeout=None, seed=seed)
            response1 = sampler.sample(bqm, num_reads=1, tenure=tenure, num_restarts=1, timeout=None, seed=seed)

            samples0 = response0.record.sample
            samples1 = response1.record.sample

            self.assertTrue(np.array_equal(samples0, samples1), "Same seed returned different results")

            for previous_sample in all_samples:
                self.assertFalse(np.array_equal(samples0, previous_sample), "Different seed returned same results")

            all_samples.append(samples0)

    def test_timeout(self):
        sampler = tabu.TabuSampler()
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': 1, 'ac': 1})

        with tictoc() as tt:
            response = sampler.sample(bqm, num_reads=1, timeout=500, seed=123)
        self.assertAlmostEqual(tt.dt, 0.5, places=1)

        with tictoc() as tt:
            response = sampler.sample(bqm, num_reads=3, timeout=200, seed=123)
        self.assertAlmostEqual(tt.dt, 0.6, places=1)

    def test_num_restarts(self):
        sampler = tabu.TabuSampler()
        bqm = dimod.generators.random.randint(10, 'SPIN', seed=123)
        target_restarts = 100

        response = sampler.sample(bqm, num_reads=1, timeout=None, num_restarts=target_restarts, seed=345)

        num_restarts = response.record['num_restarts']  
        self.assertEqual(target_restarts, num_restarts)

    def test_energy_threshold(self):
        sampler = tabu.TabuSampler()
        bqm = dimod.generators.random.randint(100, 'SPIN', seed=123)
        energy_threshold = -400

        # Expect that energy_threshold is met before timeout
        with tictoc() as tt:
            response = sampler.sample(bqm, timeout=100000, energy_threshold=energy_threshold, seed=345)

        self.assertLessEqual(tt.dt, 1.0)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the (private) TabuSearch python interface."""

import unittest
from concurrent.futures import ThreadPoolExecutor, wait

import dimod
import numpy as np

import tabu
from tabu.utils import tictoc


class TestTabuSearch(unittest.TestCase):

    def test_trivial(self):
        qubo = [[1.0]]
        init = [1]
        tenure = len(init) - 1
        timeout = 1
        restarts = 100

        search = tabu.TabuSearch(qubo, init, tenure, timeout, restarts)

        solution = list(search.bestSolution())
        energy = search.bestEnergy()

        self.assertEqual(solution, [0])
        self.assertEqual(energy, 0.0)

    def test_correctness(self):
        qubo = [[-1.2, 1.1], [1.1, -1.2]]
        init = [1, 1]
        tenure = len(init) - 1
        timeout = 20
        restarts = 100

        search = tabu.TabuSearch(qubo, init, tenure, timeout, restarts)

        solution = list(search.bestSolution())
        energy = search.bestEnergy()

        self.assertEqual(solution, [0, 1])
        self.assertEqual(energy, -1.2)

    def test_concurrency(self):

        def search(timeout, restarts=int(1e6)):
            return tabu.TabuSearch([[1.0]], [1], 0, timeout, restarts).bestEnergy()

        with ThreadPoolExecutor(max_workers=3) as executor:

            # ~ 0.5s (but be gracious on slow CI VMs)
            with tictoc() as tt:
                wait([executor.submit(search, timeout=500) for _ in range(3)])
            self.assertTrue(0.4 < tt.dt < 1.6)

            # ~ 1s (but be gracious on slow CI VMs)
            with tictoc() as tt:
                wait([executor.submit(search, timeout=500) for _ in range(4)])
            self.assertTrue(0.9 < tt.dt < 2.1)

    def test_float(self):
        n = 20
        init = [1] * n
        tenure = len(init) - 1
        timeout = 20
        restarts = 100

        bqm = dimod.generators.random.uniform(n, 'BINARY', low=-100, high=100, seed=123)
        Q, _ = tabu.TabuSampler._bqm_to_tabu_qubo(bqm)
        search = tabu.TabuSearch(Q, init, tenure, timeout, restarts)
        self.assertAlmostEqual(search.bestEnergy(), -1465.9867898)

        bqm = dimod.generators.random.uniform(n, 'BINARY', low=-1, high=1, seed=123)
        Q, _ = tabu.TabuSampler._bqm_to_tabu_qubo(bqm)
        search = tabu.TabuSearch(Q, init, tenure, timeout, restarts)
        self.assertAlmostEqual(search.bestEnergy(), -14.65986790)

    def test_exceptions(self):
        qubo = [[-1.2, 1.1], [1.1, -1.2]]
        timeout = 10
        restarts = 100

        # Wrong length for init_solution
        with self.assertRaises(RuntimeError):
            init = [1, 1, 1]
            tenure = len(init) - 1
            search = tabu.TabuSearch(qubo, init, tenure, timeout, restarts)

        # Tenure out of bounds
        with self.assertRaises(RuntimeError):
            init = [1, 1]
            tenure = 3
            search = tabu.TabuSearch(qubo, init, tenure, timeout, restarts)
